from typing import Callable

import jax
import jax.numpy as jnp
from jax import vmap, jit
from kalmax.utils import gaussian_pdf
from kalmax.kernels import gaussian_kernel

from functools import partial

def kde(
        bins: jnp.ndarray,
        trajectory: jnp.ndarray,
        spikes: jnp.ndarray,
        kernel : Callable = gaussian_kernel,
        kernel_bandwidth: float = 0.01,
        mask: jnp.ndarray = None,
        batch_size: int = 36000,
        ) -> jnp.ndarray:
    """
    Performs KDE to estimate the expected number of spikes each neuron will fire at each position in `bins` given past `trajectory` and `spikes` data. This estimate is an expected-spike-count-per-timebin, in order to get firing rate in Hz, divide this by dt.

    Kernel Density Estimation goes as follows (the denominator corrects for for non-uniform position density): 

              # spikes observed at x     sum_{spike_times} K(x, x(ts))     Ks
      mu(x) = ---------------------- ==> ----------------------------- :=  --
                  # visits to x            sum_{all_times} K(x, x(t))      Kx
              = exp[log(Ks) - log(Kx)]

    Optionally, a boolean mask same shape as spikes can be passed to ignore certain spikes. This restricts the KDE calculation to only the spikes where mask is True.
    
    Parameters
    ----------
    bins : jnp.ndarray, shape (N_bins, D,)
        The position bins at which to estimate the firing rate
    trajectory : jnp.ndarray, shape (T, D)
        The position of the agent at each time step
    spikes : jnp.ndarray, shape (T, N_neurons)
        The spike counts of the neuron at each time step (integer array, can be > 1)
    kernel : function
        The kernel function to use for density estimation. See `kernels.py` for signature and examples.
    kernel_bandwidth : float
        The bandwidth of the kernel
    mask : jnp.ndarray, shape (T, N_neurons), optional
        A boolean mask to apply to the spikes. If None, no mask is applied. Default is None.
    batch_size : int
        The time axis is split into batches of this size to avoid memory errors, each batch is then processed in series. Default is 36000 (chosen to be 1 hr at 10 and an amount which doesn't crash CPU)

    
    Returns
    -------
    kernel_density_estimate : jnp.ndarray, shape (N_neurons, N_bins)
    """
    assert bins.ndim == 2
    assert trajectory.ndim == 2
    assert spikes.ndim == 2

    N_neurons = spikes.shape[1]
    N_bins = bins.shape[0]
    T = trajectory.shape[0]

    # If not passed make a trivial mask (all True)
    if mask is None: mask = jnp.ones_like(spikes, dtype=bool)
    
    # vmap the kernel K(x,mu,sigma) so it takes in a vector of positions and a vector of means
    kernel_fn = partial(kernel, bandwidth=kernel_bandwidth)
    vmapped_kernel = vmap(vmap(kernel_fn, in_axes=(0, None)), in_axes=(None, 0))

    spike_density = jnp.zeros((N_bins, N_neurons))
    position_density = jnp.zeros((N_bins, N_neurons))

    N_batchs = int(jnp.ceil(T / batch_size))
    for i in range(N_batchs):
        start = i * batch_size
        end = min((i+1) * batch_size, T)
        
        # Get the batch of trajectory, spikes and mask
        trajectory_batch = trajectory[start:end]
        spikes_batch = spikes[start:end]
        mask_batch = mask[start:end]

        # Pairwise kernel values for each trajectory-bin position pair. The bulk of the computation is done here. 
        kernel_values = vmapped_kernel(trajectory_batch, bins)
        # Calculate normalisation position density (the +epsilon is means unvisited positions should approach 0 density and avoid nans)
        position_density_batch = kernel_values @ mask_batch + 1e-6
        # Calculate spike density, replace nans from no-spikes with 0
        spike_density_batch = kernel_values @ (mask_batch*spikes_batch)
        spike_density_batch = jnp.where(jnp.isnan(spike_density_batch), 0, spike_density_batch)

        # Add these to the running total
        spike_density += spike_density_batch
        position_density += position_density_batch

    # calculate kde at each bin position 
    kernel_density_estimate = jnp.exp(jnp.log(spike_density) - jnp.log(position_density)).T

    return kernel_density_estimate


def poisson_log_likelihood(spikes : jnp.ndarray,
                           mean_rate : jnp.ndarray,
                           mask : jnp.ndarray = None,
                           renormalise=True):
        """Takes an array of spike counts and an array of mean rates and returns the log-likelihood of the spikes given the mean rate of the neuron (it's receptive field). 

        P(X|mu) = (mu^X * e^-mu) / X!
        log(P(X|mu)) = sum_{neurons} [ X * log(mu) - mu - log(X!) ]
        where 
        log(X!) = log(sqrt(2*pi)) + (X+0.5) * log(X) - X    (manually correcting for when X=0) #this stirlings approximation IS necessary as it avoids going through n! which can be enormous and give nans for large spike counts 

        Optionally, a boolean mask same shape as spikes can be passed to ignore certain spikes. This restricts the likelihood calculation to only the spikes where mask is True.
        
        Parameters
        ----------
        spikes : jnp.ndarray, shape (T, N_neurons,)
            How many spikes the neuron actually fired at each bin (int, can be > 1)
        mean_rate : jnp.ndarray, shape (N_neurons, N_bins,)
            The mean rate of the neuron (it's receptive field) at each bin. This is how many spikes you would _expect_ in at this position in a time dt.
        mask : jnp.ndarray, shape (T, N_neurons,), optional
            A boolean mask to apply to the spikes. If None, no mask is applied. Default is None.
        renormalise : bool, optional
            If True this renormalises so the maximum log-likelihood is always 0 (max likelihood is 1). Recommended to avoid nan errors when likelihoods are small. Default is True.
            
        Returns
        -------
        log_likelihood : jnp.ndarray, shape (T, N_bins,)
            The log-likelihood (summed over neurons) of the spikes given the mean rate of the neuron
        """
        # If not passed make a no-mask mask (all True)
        if mask is None: mask = jnp.ones_like(spikes, dtype=bool)

        # Calculate log factorial of spike counts NOTE this could be removed if you dont care about absolute likelihoods
        spikes_ = jnp.where(spikes == 0, 1, spikes) # replace 0 spikes with 1s because 0! = 1
        log_spikecount_factorial = jnp.log(jnp.sqrt(2*jnp.pi)) + (spikes_ + 0.5) * jnp.log(spikes_) - spikes_ # manually correcting for when X=0
        
        # Sum over neurons (which are unmasked) 
        logPXmu = (mask*spikes) @ jnp.log(mean_rate+1e-3) - mask @ mean_rate - jnp.sum(log_spikecount_factorial * mask, axis=1)[:,None]
        
        # Renormalise so max likelihood is 1
        if renormalise: logPXmu = logPXmu - jnp.max(logPXmu, axis=1)[:,None]
        return logPXmu


def poisson_log_likelihood_trajectory(spikes : jnp.ndarray,
                                      mean_rate_along_trajectory : jnp.ndarray,
                                      mask : jnp.ndarray = None,):
    """Takes an array of spike counts and an _equally shaped_ array of mean rates and returns the log-likelihood of the spikes given the mean rate of the neuron (it's receptive field). This is different from `poisson_log_likelihood` in that it takes in a trajectory of mean rates and spikes and returns the log-likelihood of the spikes given the trajectory of mean rates. 

    Parameters
    ----------
    spikes : jnp.ndarray, shape (T, N_neurons,)
        How many spikes the neuron actually fired at each bin (int, can be > 1)
    mean_rate_along_trajectory : jnp.ndarray, shape (T, N_neurons,)
        The mean rate of the neurons as calculated at each time step along the trajectory. This is how many spikes you would _expect_ in at this position in a time dt.
    mask : jnp.ndarray, shape (T, N_neurons,), optional
        A boolean mask to apply to the spikes. If None, no mask is applied. Default is None.
    renormalise : bool, optional
        If True this renormalises so the maximum log-likelihood is always 0 (max likelihood is 1). Recommended to avoid nan errors when likelihoods are small. Default is True.

    Returns
    -------
    log_likelihood : jnp.ndarray, shape (T,)
        The log-likelihood (summed over neurons) of the spikes given the mean rate of the neuron
    """
        
    # If not passed make a no-mask mask (all True)
    if mask is None: mask = jnp.ones_like(spikes, dtype=bool)
      
    # Calculate log factorial of spike counts NOTE this could be removed, its just a constant factor
    spikes_ = jnp.where(spikes == 0, 1, spikes) # replace 0 spikes with 1s because 0! = 1
    log_spikecount_factorial = jnp.log(jnp.sqrt(2*jnp.pi)) + (spikes_ + 0.5) * jnp.log(spikes_) - spikes_ # manually correcting for when X=0
      
    # Calculate log-likelihood and sum over (unmasked) neurons
    logPXmu = (spikes*jnp.log(mean_rate_along_trajectory+1e-3)) - (mean_rate_along_trajectory) -  (log_spikecount_factorial)
    logPXmu = jnp.sum(mask * logPXmu, axis=1)

    return logPXmu
      




import jax 
import jax.numpy as jnp
from jax import jit, vmap
import math
from kalmax.utils import gaussian_pdf, log_gaussian_pdf

# TODO check that jnp.einsum isn't slowing things down

class KalmanFilter:
    """A Kalman filter class. This class is used to filter the data and fit the model.
    Written in jax, the lower level functions are jit compiled for speed. The filtering and smoothing loops are processed in batches using jax.lax.scan(): higher batch sizes will run faster but at the cost of a one-off compilation time.
    
    The Kalman dynamics equations are as follows:
    z_t = F @ z_t-1 + B @ u_t + q_t
    y_t = H @ z_t + r_t
    where z_t is the hidden state, y_t is the observation, u_t is the control input, F is the state transition matrix, B is the control matrix, H is the observation matrix, q_t ~ N(0, Q) is the state transition noise, and r_t ~ N(0, R) is the observation noise.

    Kalman _filtering_ takes observations and estimates the _causal_ posterior distribution of the hidden state given the observations. Kalman _smoothing_ takes the filtered estimates and estimates the _posterior_ distribution of the hidden state given all the observations.
    mu_filter_t = E[z_t | y_1:t, u_1:t]
    sigma_filter_t = Cov[z_t | y_1:t, u_1:t]
    mu_smooth_t = E[z_t | y_1:T, u_1:T]
    sigma_smooth_t = Cov[z_t | y_1:T, u_1:T]
    """

    def __init__(self, dim_Z : int, 
                       dim_Y : int,
                       dim_U : int = 0,
                       batch_size : int = 100, 
                       # optional parameters
                       mu0 : jnp.ndarray = None,
                       sigma0 : jnp.ndarray = None,
                       F : jnp.ndarray = None,
                       B : jnp.ndarray = None,
                       Q : jnp.ndarray = None,
                       H : jnp.ndarray = None,
                       R : jnp.ndarray = None,
                       ):
        """Initializes the Kalman class. The state has size dim_Z, the observations have size dim_Y, and the control input has size dim_U.
        
        If dim_U = 0, no control input is used.
        
        Parameters F, B, Q, H and R can either be:
        * Passed in at initialisation --> assumed constant over time
        * Passed in at runtime --> assumed to time-vary (additional time-dim in along the 0 axis matching the length of the observation data)

        Parameters
        ----------
        dim_Z : int
            The size of the state space
        dim_Y : int
            The size of the observation space
        dim_U : int, optional
            The size of the control space (default is 0, for no control)
        batch_size : int
            The batch size for the Kalman filter

        Optional parameters
        -------------------
        mu0 : jnp.ndarray, shape (dim_Z,)
            The initial state mean
        sigma0 : jnp.ndarray, shape (dim_Z, dim_Z)
            The initial state covariance
        F : jnp.ndarray, shape (dim_Z, dim_Z)
            The state transition matrix
        B : jnp.ndarray, shape (dim_Z, dim_U)
            The control matrix
        Q : jnp.ndarray, shape (dim_Z, dim_Z)
            The state transition noise covariance
        H : jnp.ndarray, shape (dim_X, dim_Z)
            The observation matrix
        R : jnp.ndarray, shape (dim_X, dim_X)
            The observation noise covariance
        """
            
        self.dim_Z = dim_Z
        self.dim_Y = dim_Y
        self.dim_U = dim_U
        self.batch_size = batch_size
        
        # Optionally set parameters and initial conditions now
        if mu0 is not None:
            assert mu0.shape == (self.dim_Z,)
        if sigma0 is not None:
            assert sigma0.shape == (self.dim_Z, self.dim_Z)
        if F is not None:
            assert F.shape == (self.dim_Z, self.dim_Z)
        if B is not None:
            assert B.shape == (self.dim_Z, self.dim_U)  
        if Q is not None:
            assert Q.shape == (self.dim_Z, self.dim_Z)
        if H is not None:
            assert H.shape == (self.dim_Y, self.dim_Z)
        if R is not None:
            assert R.shape == (self.dim_Y, self.dim_Y)
    
        self.mu0 = mu0
        self.sigma0 = sigma0
        self.F = F
        self.B = B if B is not None else jnp.zeros((self.dim_Z, self.dim_U))
        self.Q = Q
        self.H = H
        self.R = R

    def filter(self, Y, 
                     U = None, # <-- MODIFIED: Made optional
                     mu0=None, 
                     sigma0=None, 
                     F=None,
                     B=None, # <-- ADDED
                     Q=None,
                     H=None,
                     R=None,):
        """Takes sequences of observations, control inputs, and noise covariances and runs the Kalman filter on the data. If parameters are not passed in, the class defaults are used. If they are passed in, they must have shape (T, *param_shape,) where T is the number of time steps - this allows for time-varying parameters.
        
        Parameters
        ----------
        Y : jnp.ndarray, shape (T, dim_Y)
            The observation means
        U : jnp.ndarray, shape (T, dim_U), optional
            The control inputs (defaults to zeros if not provided)
        mu0 : jnp.ndarray, shape (dim_Z,)
            The initial state mean, optional (default is provided at initialisation)
        sigma0 : jnp.ndarray, shape (dim_Z, dim_Z)
            The initial state covariance, optional (default is provided at initialisation)
        F : jnp.ndarray, shape (T, dim_Z, dim_Z)
            The state transition matrix, optional (default is provided at initialisation)
        B : jnp.ndarray, shape (T, dim_Z, dim_U)
            The control matrix, optional (default is provided at initialisation)
        Q : jnp.ndarray, shape (T, dim_Z, dim_Z)
            The state transition noise covariance, optional (default is provided at initialisation)
        H : jnp.ndarray, shape (T, dim_Z, dim_Z)
            The observation matrix, optional (default is provided at initialisation)
        R : jnp.ndarray, shape (T, dim_Z, dim_Z)
            The observation noise covariances, optional (default is provided at initialisation)


        Returns
        -------
        mus_f : jnp.ndarray, shape (T, dim_Z)
            The filtered means
        sigmas_f : jnp.ndarray, shape (T, dim_Z, dim_Z)
            The filtered covariances
        """
        assert Y.ndim == 2; assert Y.shape[1] == self.dim_Y
        T = Y.shape[0] # number of time steps

        if mu0 is None: 
            assert self.mu0 is not None, "You must either pass in the initial conditions or set them at initialisation"
            mu0 = self.mu0
        else: 
            assert mu0.ndim == 1; assert mu0.shape[0] == self.dim_Z
        if sigma0 is None:
            assert self.sigma0 is not None, "You must either pass in the initial conditions or set them at initialisation"
            sigma0 = self.sigma0
        else:
            assert sigma0.ndim == 2; assert sigma0.shape[0] == self.dim_Z; assert sigma0.shape[1] == self.dim_Z
            
        F = self._verify_and_tile(F, self.F, T, (self.dim_Z, self.dim_Z))
        B = self._verify_and_tile(B, self.B, T, (self.dim_Z, self.dim_U))
        Q = self._verify_and_tile(Q, self.Q, T, (self.dim_Z, self.dim_Z))
        H = self._verify_and_tile(H, self.H, T, (self.dim_Y, self.dim_Z))
        R = self._verify_and_tile(R, self.R, T, (self.dim_Y, self.dim_Y))

        if U is None:
            U = jnp.zeros((T, self.dim_U))
        else:
            assert U.ndim == 2; assert U.shape[0] == T; assert U.shape[1] == self.dim_U

        mus_f, sigmas_f = [],[] # filtered means and covariances
        
        N_batch = math.ceil(T/self.batch_size)
        for i in range(N_batch):
            start = i*self.batch_size
            end = min((i+1)*self.batch_size, T)
            mu, sigma = kalman_filter(
                    Y = Y[start:end],
                    U = U[start:end],
                    mu0 = mu0, 
                    sigma0 = sigma0, 
                    F = F[start:end], 
                    B = B[start:end],
                    Q = Q[start:end], 
                    H = H[start:end], 
                    R = R[start:end], )
            mus_f.append(mu)
            sigmas_f.append(sigma)
            mu0, sigma0 = mu[-1], sigma[-1] # update initial conditions for next batch
        mus_f = jnp.concatenate(mus_f)
        sigmas_f = jnp.concatenate(sigmas_f)

        return mus_f, sigmas_f
    
    def smooth(self, mus_f, 
                     sigmas_f, 
                     U = None,
                     F=None, 
                     B=None,
                     Q=None,): 
        """Takes the filtered means, covariances, and control inputs and runs the Kalman smoother on the data.

        Parameters
        ----------
        mus_f : jnp.ndarray, shape (T, dim_Z)
            The filtered means
        sigmas_f : jnp.ndarray, shape (T, dim_Z, dim_Z)
            The filtered covariances
        U : jnp.ndarray, shape (T, dim_U), optional
            The control inputs (defaults to zeros if not provided)
        F : jnp.ndarray, shape (T, dim_Z, dim_Z)
            The state transition matrix, optional
        B : jnp.ndarray, shape (T, dim_Z, dim_U)
            The control matrix, optional
        Q : jnp.ndarray, shape (T, dim_Z, dim_Z)
            The state transition noise covariance, optional
        
        Returns
        -------
        mus_s : jnp.ndarray, shape (T, dim_Z)
            The smoothed means
        sigmas_s : jnp.ndarray, shape (T, dim_Z, dim_Z)
            The smoothed covariances
        """

        T = len(mus_f)
        muT = mus_f[-1]
        sigmaT = sigmas_f[-1]
        mus_s, sigmas_s = [jnp.array([muT])],[jnp.array([sigmaT])]

        F = self._verify_and_tile(F, self.F, T, (self.dim_Z, self.dim_Z))
        B = self._verify_and_tile(B, self.B, T, (self.dim_Z, self.dim_U))
        Q = self._verify_and_tile(Q, self.Q, T, (self.dim_Z, self.dim_Z))

        if U is None:
            U = jnp.zeros((T, self.dim_U))
        else:
            assert U.ndim == 2; assert U.shape[0] == T; assert U.shape[1] == self.dim_U

        for i in range(math.ceil((T-1)/(self.batch_size))):
            start = max(0, T - 1 - (i+1)*self.batch_size)
            end = T - 1 - i*self.batch_size
            mu, sigma = kalman_smoother(
                mu = mus_f[start:end], 
                sigma = sigmas_f[start:end], 
                U = U[start:end],
                muT = muT, 
                sigmaT = sigmaT, 
                F = F[start:end],
                B = B[start:end],
                Q = Q[start:end],)
            mus_s.insert(0,mu); sigmas_s.insert(0,sigma)
            muT, sigmaT = mu[0], sigma[0]
        mus_s = jnp.concatenate(mus_s)
        sigmas_s = jnp.concatenate(sigmas_s)

        return mus_s, sigmas_s

    def loglikelihood(self, Y,
                            mu, 
                            sigma, 
                            H=None, 
                            R=None,):
        """Calculates the log-likelihood of the observations, Y, by marginalising over the hidden state [mu, sigma] (filtered or smoothed). This can be done analytically (see page 361 of the Advanced Murphy book). 

        P(Y) = Normal(Y | Y_hat, S) where S = H @ sigma @ H.T + R (the posterior observation covariance combined with the observation noise covariance) Y_hat = H @ mu (the predicted observation). 

        Parameters
        ----------
        Y : jnp.ndarray, shape (T, dim_Y)
            The observation means
        mu : jnp.ndarray, shape (T, dim_Z)
            The posterior state means
        sigma : jnp.ndarray, shape (T, dim_Z, dim_Z)
            The posterior state covariances
        H: jnp.ndarray, shape (T, dim_Y, dim_Z)
            The observation matrix, optional
        R : jnp.ndarray, shape (T, dim_Y, dim_Y)
            The observation noise covariances, optional

        Returns
        -------
        logP : jnp.ndarray, shape (T,)
            The log-likelihood of the data given the model
        """

        T = len(mu)
        H = self._verify_and_tile(H, self.H, T, (self.dim_Y, self.dim_Z))
        R = self._verify_and_tile(R, self.R, T, (self.dim_Y, self.dim_Y))
            
        S = vmap(calculate_S_matrix, (0, 0, 0))(sigma, H, R)
        Y_hat = jnp.einsum('ijk,ik->ij', H, mu) # the "observation" mean
        logP = vmap(log_gaussian_pdf, (0, 0, 0))(Y, Y_hat, S)
        # logP = jnp.log(P) # TODO multiply this by a volume element to get actual probability

        return logP

    def _verify_and_tile(self, param, default_param, T, intended_shape):
        """Verifies the shape of the parametery. If the parameter is not passed in, the default parameter (presumably set at initialisation) is tiled T-times along a new 0th axis and used. 

        Parameters
        ----------
        param : jnp.ndarray or None
            The parameter to be verified
        default_param : jnp.ndarray
            The default parameter
        T : int
            The number of times to tile the parameter
        intended_shape : tuple
            The intended shape of the parameter

        Returns
        -------
        param : jnp.ndarray
            The parameter, shape (T, *intended_shape)
        """
        if param is None: 
            assert default_param is not None, "You must either pass in the parameter or set it at initialisation"
            param = jnp.tile(default_param, (T,)+(1,)*default_param.ndim)
        else: 
            assert param.shape == (T,)+intended_shape, f"Parameter shape is {param.shape} but should be {(T,)+intended_shape}"
        return param

@jit 
def kalman_filter(Y, U, mu0, sigma0, F, B, Q, H, R):
    """Kalman filters a batch of observation data, Y. 

    Parameters
    ----------
    Y : jnp.ndarray, shape (T, dim_Y)
        The observation means
    U : jnp.ndarray, shape (T, dim_U)
        The control inputs
    mu0 : jnp.ndarray, shape (dim_Z,)
        The initial state mean
    sigma0 : jnp.ndarray, shape (dim_Z, dim_Z)
        The initial state covariance
    F : jnp.ndarray, shape (T, dim_Z, dim_Z)
        The state transition matrix
    B : jnp.ndarray, shape (T, dim_Z, dim_U)
        The control matrix
    Q : jnp.ndarray, shape (T, dim_Z, dim_Z)
        The state transition noise covariance
    H : jnp.ndarray, shape (T, dim_Y, dim_Z)
        The observation matrix
    R : jnp.ndarray, shape (T, dim_Y, dim_Y)
        The observation noise covariances


    Returns
    -------
    mu : jnp.ndarray, shape (T, dim_Z)
        The filtered posterior state means
    sigma : jnp.ndarray, shape (T, dim_Z, dim_Z)
        The filtered posterior state covariances
    """
    def loop(carry, inputs):
        mu, sigma = carry
        Y, u, F, B, Q, H, R, = inputs
        mu_p, sigma_p = kalman_predict(mu, sigma, F, Q, B, u)
        mu_u, sigma_u = kalman_update(mu_p, sigma_p, H, R, Y)
        return (mu_u, sigma_u), (mu_u, sigma_u) # carry, output
    _, (mu_all, sigma_all) = jax.lax.scan(loop, (mu0, sigma0), (Y, U, F, B, Q, H, R))
    return jnp.stack(mu_all), jnp.stack(sigma_all)

@jit
def kalman_smoother(mu, sigma, U, muT, sigmaT, F, B, Q): # <-- ADDED U, B
    """Runs the Kalman smoother on the data. mu and sigma are in forward order, ie. mu = [mu[0], mu[1], ... mu[T]] and they are looped over in reverse order, so you can still batch the data. 


    Parameters
    ----------
    mu : jnp.ndarray, shape (T, dim_Z)
        The filtered posterior state means 
    sigma : jnp.ndarray, shape (T, dim_Z, dim_Z)
        The filtered posterior state covariances
    U : jnp.ndarray, shape (T, dim_U)
        The control inputs
    muT : jnp.ndarray, shape (dim_Z,)
        The final state mean - by definition this should have already been smoothed
    sigmaT : jnp.ndarray, shape (dim_Z, dim_Z)
        The final state covariance - by definition this should have already been smoothed
    F : jnp.ndarray, shape (T, dim_Z, dim_Z)
        The state transition matrix
    B : jnp.ndarray, shape (T, dim_Z, dim_U)
        The control matrix
    Q : jnp.ndarray, shape (T, dim_Z, dim_Z)
        The state transition noise covariance

    Returns
    -------
    mu_smooth : jnp.ndarray, shape (T, dim_Z)
        The smoothed state means
    sigma_smooth : jnp.ndarray, shape (T, dim_Z, dim_Z)
        The smoothed state covariances
    """
    def loop(carry, inputs):
        mu, sigma = carry
        mu_, sigma_, u, F, B, Q = inputs
        mu_predict, sigma_predict = kalman_predict(mu_, sigma_, F, Q, B, u)
        J = sigma_ @ F.T @ jnp.linalg.inv(sigma_predict)
        mu_smoothed = mu_ + J @ (mu - mu_predict)
        sigma_smoothed = sigma_ + J @ (sigma - sigma_predict) @ J.T
        return (mu_smoothed, sigma_smoothed), (mu_smoothed, sigma_smoothed)
    _, (mus_all, sigmas_all) = jax.lax.scan(loop, (muT, sigmaT), (mu[::-1], sigma[::-1], U[::-1], F[::-1], B[::-1], Q[::-1]))
    mus_all = mus_all[::-1]
    sigmas_all = sigmas_all[::-1] # reverse the order back to forward

    return mus_all, sigmas_all

@jit
def kalman_likelihoods(Z, Y, mu, sigma, F, Q, H, R, B=None, U=None): # <-- MODIFIED
    """Calculates the prior P(Z), likelihood P(Y | Z), and posterior P(Z | Y) of any state trajectory (Z) and observations (Y, R) under the fitted kalman model. Note although Z and Y can, in principle, be _any_ trajectory and observations, typically Z == mu and Y == the observations which were used to fit the model in the first place. 

    Parameters
    ----------
    Z : jnp.ndarray, shape (T, dim_Z)
        The trajectory of the agent (typical this might just be the same as mu) 
    Y : jnp.ndarray, shape (T, dim_Y)
        The observations to be evalauted 
    mu : jnp.ndarray, shape (T, dim_Z)
        The posterior state means 
    sigma : jnp.ndarray, shape (T, dim_Z, dim_Z)
        The posterior state covariances
    F : jnp.ndarray, shape (T, dim_Z, dim_Z)
        The state transition matrix
    Q : jnp.ndarray, shape (T, dim_Z, dim_Z)
        The state transition noise covariance
    H : jnp.ndarray, shape (T, dim_Y, dim_Z)
        The observation matrix
    R : jnp.ndarray, shape (T, dim_Y, dim_Y)
        The observation noise covariances
    B : jnp.ndarray, shape (T, dim_Z, dim_U), optional
        The control matrix
    U : jnp.ndarray, shape (T, dim_U), optional
        The control inputs

    Returns
    -------
    PZ : jnp.ndarray, shape (T,)
        The likelihood of the state given the previous state
    PZXF : jnp.ndarray, shape (T,)
        The likelihood of the state given the observation
    PXZF : jnp.ndarray, shape (T,)
        The likelihood of the observation given the state
    """
    
    T = Z.shape[0]
    dim_Z = Z.shape[1]
    if B is None:
        dim_U = 0
        B = jnp.zeros((T, dim_Z, dim_U))
        U = jnp.zeros((T, dim_U))
    elif U is None:
        dim_U = B.shape[-1]
        U = jnp.zeros((T, dim_U))

    Z_ = jnp.concatenate((Z[0][None], Z)) #prepend Z0 to Z so its [Z0, Z0, Z1, Z2, ... ZT]
    U_ = jnp.concatenate((U[0][None], U)) #prepend U0 to U so its [U0, U0, U1, U2, ... UT]
    Q_ = jnp.concatenate((Q[0][None], Q)) #prepend Q0 to Q so its [Q0, Q0, Q1, Q2, ... QT]
    F_ = jnp.concatenate((F[0][None], F)) #prepend F0 to F so its [F0, F0, F1, F2, ... FT]
    B_ = jnp.concatenate((B[0][None], B)) #prepend B0 to B so its [B0, B0, B1, B2, ... BT]
    
    mu_p = jnp.einsum('ijk,ik->ij', F_, Z_) + jnp.einsum('ijk,ik->ij', B_, U_)
    
    Y_hat = jnp.einsum('ijk,ik->ij', H, mu) # the "observation" mean
    PZ   = vmap(gaussian_pdf,(0,0,0))(Z_[1:], mu_p[:-1], Q_[1:]) # zt ~ N(F*zt-1 + B*ut, Qt)
    PZXF = vmap(gaussian_pdf,(0,0,0))(Z, mu, sigma) # zt ~ N(mu, sigma)
    PXZF = vmap(gaussian_pdf,(0,0,0))(Y, Y_hat, R)
    return PZ, PZXF, PXZF

def kalman_predict(mu, sigma, F, Q, B, u):
    """Predicts the next state of the system given the current state and the state transition matrix.

    Parameters
    ----------
    mu : jnp.ndarray, shape (dim_Z,)
        The current state mean
    sigma : jnp.ndarray, shape (dim_Z, dim_Z)
        The current state covariance
    F : jnp.ndarray, shape (dim_Z, dim_Z)
        The state transition matrix
    Q : jnp.ndarray, shape (dim_Z, dim_Z)
        The state transition noise covariance
    B : jnp.ndarray, shape (dim_Z, dim_U)
        The control matrix
    u : jnp.ndarray, shape (dim_U,)
        The control input

    Returns
    -------
    mu_next : jnp.ndarray, shape (dim_Z,)
        The predicted next state mean
    sigma_next : jnp.ndarray, shape (dim_Z, dim_Z)
        The predicted next state covariance
    """
    mu_next = F @ mu + B @ u
    sigma_next = F @ sigma @ F.T + Q
    return mu_next, sigma_next

def kalman_update(mu, sigma, H, R, y):
    """Updates the state estimate given an observation.

    Parameters
    ----------
    mu : jnp.ndarray, shape (dim_Z,)
        The current state mean
    sigma : jnp.ndarray, shape (dim_Z, dim_Z)
        The current state covariance
    H : jnp.ndarray, shape (dim_Y, dim_Z)
        The observation matrix
    R : jnp.ndarray, shape (dim_Y, dim_Y)
        The observation noise covariance
    y : jnp.ndarray, shape (dim_Y,)
        The state observation

    Returns
    -------
    mu_post : jnp.ndarray, shape (dim_Z,)
        The posterior state mean
    sigma_post : jnp.ndarray, shape (dim_Z, dim_Z)
        The posterior state covariance
    """
    S = calculate_S_matrix(sigma, H, R) 
    y_hat = H @ mu
    K = calculate_K_matrix(sigma, H, S)
    mu_post = mu + K @ (y - y_hat)
    sigma_post = (jnp.eye(len(mu)) - K @ H) @ sigma

    return mu_post, sigma_post

def calculate_S_matrix(sigma, H, R):
    """Calculates the S matrix for the Kalman filter. This doesn't really need to be it's own function but it's useful for readability and I vmap it later.
    
    Parameters
    ----------
    sigma : jnp.ndarray, shape (dim_Z, dim_Z)
        The state covariance
    H : jnp.ndarray, shape (dim_Y, dim_Z)
        The observation matrix
    R : jnp.ndarray, shape (dim_Y, dim_Y)
        The observation noise covariance
        
    Returns
    -------
    S : jnp.ndarray, shape (dim_Y, dim_Y)
        The S matrix    
    """
    return H @ sigma @ H.T + R

def calculate_K_matrix(sigma, H, S):
    """Calculates the K matrix for the Kalman filter. This doesn't really need to be it's own function but it's useful for readability and I vmap it later.

    Parameters
    ----------
    sigma : jnp.ndarray, shape (dim_Z, dim_Z)
        The state covariance
    H : jnp.ndarray, shape (dim_Y, dim_Z)
        The observation matrix
    S : jnp.ndarray, shape (dim_Y, dim_Y)
        The S matrix

    Returns
    -------
    K : jnp.ndarray, shape (dim_Z, dim_Y)
        The K matrix
    """
    return sigma @ H.T @ jnp.linalg.inv(S)

def fit_parameters(Z, Y): 
    """Assuming a training set exists where hidden states Z and observations Y are known, this function fits the optimal stationary parameters of the Kalman filter, i.e. returning those that maximise the likelihood of the data and the state: L(Θ) = log({z},{y} | Θ). These solutions are (relatively) easy to derive, I took them from Byron Yu's lecture notes (they look a lot like linear regression solutions): 
    
    **NOTE: This function assumes NO control input (B=0).** Fitting B would require U as an input and a different regression setup (e.g., regressing z_t+1 on [z_t, u_t]).
    
    mu0 = (1/T) Σ{zt}
    sigma0 = (1/T) Σ{zt - mu0}{zt - mu0}.T
    F = Σ{zt+1 @ zt.T} Σ{zt zt.T}^-1
    Q = (1/T-1) Σ{zt - F @ zt-1}{zt - F @ zt-1}.T 
    H = Σ{yt @ zt.T} Σ{zt zt.T}^-1
    R = (1/T) Σ{yt - H @ zt}{yt - H @ zt}.T

    Parameters
    ----------
    Z : jnp.ndarray, shape (T, dim_Z)
        The hidden states (training data) 
    Y : jnp.ndarray, shape (T, dim_Y)
        The observations (training data)

    Returns
    -------
    mu0 : jnp.ndarray, shape (dim_Z,)
        The initial state mean
    sigma0 : jnp.ndarray, shape (dim_Z, dim_Z)
        The initial state covariance
    F : jnp.ndarray, shape (dim_Z, dim_Z)
        The state transition matrix
    Q : jnp.ndarray, shape (dim_Z, dim_Z)
        The state transition noise covariance
    H : jnp.ndarray, shape (dim_Y, dim_Z)
        The observation matrix
    R : jnp.ndarray, shape (dim_Y, dim_Y)
        The observation noise covariance

    """

    assert Z.ndim == 2; assert Y.ndim == 2    
    T = Z.shape[0]
    assert Y.shape[0] == T

    mu0 = Z.mean(axis=0)
    sigma0 = (1 / T) * ((Z - mu0).T @ (Z - mu0))
    F = (Z[1:].T @ Z[:-1]) @ jnp.linalg.inv(Z.T @ Z)
    Q = (1 / (T-1)) * (Z[1:] - Z[:-1] @ F.T).T @ (Z[1:] - Z[:-1] @ F.T)
    H = (Y.T @ Z) @ jnp.linalg.inv(Z.T @ Z)
    R = (1 / T) * (Y - Z @ H.T).T @ (Y - Z @ H.T)

    return mu0, sigma0, F, Q, H, R

def fit_mu0(Z):
    """Fits the initial state mean of the Kalman filter under the assumption of stationary dynamics, see `fit_parameters` for more details.

    Parameters
    ----------
    Z : jnp.ndarray, shape (T, dim_Z)
        The hidden states (training data) 

    Returns
    -------
    mu0 : jnp.ndarray, shape (dim_Z,)
        The initial state mean
    """
    return Z.mean(axis=0)

def fit_sigma0(Z):
    """Fits the initial state covariance of the Kalman filter under the assumption of stationary dynamics, see `fit_parameters` for more details.

    Parameters
    ----------
    Z : jnp.ndarray, shape (T, dim_Z)
        The hidden states (training data) 

    Returns
    -------
    sigma0 : jnp.ndarray, shape (dim_Z, dim_Z)
        The initial state covariance
    """
    T = Z.shape[0]
    mu0 = Z.mean(axis=0)
    return (1 / T) * ((Z - mu0).T @ (Z - mu0))

def fit_F(Z):
    """Fits the state transition matrix of the Kalman filter under the assumption of stationary dynamics **and no control input**, see `fit_parameters` for more details.

    Parameters
    ----------
    Z : jnp.ndarray, shape (T, dim_Z)
        The hidden states (training data) 

    Returns
    -------
    F : jnp.ndarray, shape (dim_Z, dim_Z)
        The state transition matrix
    """
    return (Z[1:].T @ Z[:-1]) @ jnp.linalg.inv(Z.T @ Z)

def fit_Q(Z):
    """Fits the state transition noise covariance of the Kalman filter under the assumption of stationary dynamics **and no control input**, see `fit_parameters` for more details.

    Parameters
    ----------
    Z : jnp.ndarray, shape (T, dim_Z)
        The hidden states (training data) 

    Returns
    -------
    Q : jnp.ndarray, shape (dim_Z, dim_Z)
        The state transition noise covariance
    """
    T = Z.shape[0]
    F = (Z[1:].T @ Z[:-1]) @ jnp.linalg.inv(Z.T @ Z)
    return (1 / (T-1)) * (Z[1:] - Z[:-1] @ F.T).T @ (Z[1:] - Z[:-1] @ F.T)

def fit_H(Z, Y):
    """Fits the observation matrix of the Kalman filter under the assumption of stationary dynamics, see `fit_parameters` for more details.

    Parameters
    ----------
    Z : jnp.ndarray, shape (T, dim_Z)
        The hidden states (training data) 
    Y : jnp.ndarray, shape (T, dim_Y)
        The observations (training data)

    Returns
    -------
    H : jnp.ndarray, shape (dim_Y, dim_Z)
        The observation matrix
    """
    return (Y.T @ Z) @ jnp.linalg.inv(Z.T @ Z)

def fit_R(Z, Y):
    """Fits the observation noise covariance of the Kalman filter under the assumption of stationary dynamics, see `fit_parameters` for more details.

    Parameters
    ----------
    Z : jnp.ndarray, shape (T, dim_Z)
        The hidden states (training data) 
    Y : jnp.ndarray, shape (T, dim_Y)
        The observations (training data)

    Returns
    -------
    R : jnp.ndarray, shape (dim_Y, dim_Y)
        The observation noise covariance
    """
    T = Z.shape[0]
    H = fit_H(Z,Y)
    return (1 / T) * (Y - Z @ H.T).T @ (Y - Z @ H.T)


# Some (mostly plotting and poorly written) utilities for the Kalman filter demo
import matplotlib
import matplotlib.pyplot as plt
import numpy as np

def animate_over_time(
            plot_function,
            t_start,
            t_end,
            speed_up=10, #relative to real time
            fps=20,):
    """Animates the plot function over time_range.

    Params
    ------
    plot_function: function, it must act as it's own init function when ax=None
        Function that takes ax, time and time_start as input and returns ax.
        Must have the following signature 
        def plot_function(ax,
                            time,
                            time_start=0 # time to start (relevant if plotting a trajectory) 
                            ):
            # do something
            return ax
        _type_: _description_
    """
    # plot function should take ax and time as input
    time_per_frame = speed_up/fps
    times = np.arange(t_start, t_end, time_per_frame)
    
    def update(time, ax, fig):
        for ax_ in fig.get_axes():
            ax_.clear()
        ax = plot_function(ax=ax, t_start=t_start, t_end=time)
        return ax
    
    ax = plot_function()
    from matplotlib.animation import FuncAnimation
    anim = FuncAnimation(plt.gcf(), update, frames=times, fargs=(ax, plt.gcf(),), interval=1000/fps)        
    plt.close()
    return anim

def plot_trajectory(
        trajectory,
        time_stamps,
        ax=None, 
        t_start=None, 
        t_end=None,
        **plot_kwargs):
    """Plots a trajectory over a given time range

    Params
    ------
    ax: matplotlib axis, optional
        Axis to plot on
    t_start: float, optional
        Start time
    t_end: float, optional
        End time
    trajectory: np.ndarray, optional
        Trajectory to plot, shape (T, 2)
    time_stamps: np.ndarray, optional
        Time stamps for the trajectory, shape (T,)
    plot_kwargs: dict, optional
        Additional plotting arguments like 'color', 'alpha', 'label', 'scatter_points', 'show_line', 'linewidth', 'title', 'xlabel', 'ylabel'

    Returns
    -------
    ax: matplotlib axis
    """
    if t_start is None: t_start = time_stamps[0]
    if t_end is None: t_end = time_stamps[-1]
    if ax is None: 
        fig, ax = plt.subplots(1, 1, figsize=(6, 6)) 
    
    id_start, id_end = np.argmin(np.abs(time_stamps - t_start)), np.argmin(np.abs(time_stamps-t_end))
    trajectory_ = trajectory[id_start:id_end]

    # Get plot kwargs
    color = plot_kwargs.get('color', 'k')
    scatter_points = plot_kwargs.get('scatter_points', True)
    show_line = plot_kwargs.get('show_line', True)
    linewidth = plot_kwargs.get('linewidth',1)
    title = plot_kwargs.get('title',None)
    xlabel = plot_kwargs.get('xlabel','x [m]')
    ylabel = plot_kwargs.get('ylabel','y [m]')
    alpha = plot_kwargs.get('alpha',1)
    label = plot_kwargs.get('label',None)
    min_x = plot_kwargs.get('min_x',trajectory[:,0].min().round(1))
    max_x = plot_kwargs.get('max_x',trajectory[:,0].max().round(1))
    min_y = plot_kwargs.get('min_y',trajectory[:,1].min().round(1))
    max_y = plot_kwargs.get('max_y',trajectory[:,1].max().round(1))
    
    if show_line:
        ax.plot(trajectory_[:,0],trajectory_[:,1],color=color, linewidth=linewidth, alpha=alpha, label=label)
    if scatter_points:
        ax.scatter(trajectory_[:,0],trajectory_[:,1],color=color, linewidth=0, s=6, alpha=alpha)
    ax.set_xlim(min_x, max_x); ax.set_ylim(min_y, max_y); ax.set_aspect('equal', 'box')
    ax.set_xticks([min_x, max_x]); ax.set_yticks([min_y, max_y])
    ax.set_xlabel(xlabel)
    ax.set_ylabel(ylabel)
    if title is not None: ax.set_title(title)

    return ax 

def plot_ellipse(ax, mean, cov, color):
    lambda_, v = np.linalg.eig(cov)
    lambda_ = np.sqrt(lambda_) # convert from variance to standard deviation along the eigenvectors
    ell = matplotlib.patches.Ellipse(xy=mean,
                                     width=lambda_[0]*2, 
                                     height=lambda_[1]*2,
                                     angle=np.rad2deg(np.arctan(v[:, 0][1] / v[:, 0][0])),
                                     lw=1, 
                                     fill=True, 
                                     edgecolor=color,
                                     facecolor=color,
                                     alpha=0.5,)
    ax.add_artist(ell)
    return ax 


"""
The kernel functions to use for density estimation. All kernel functions _must_ have the following signature: 
    
    `kernel(x_1: jnp.ndarray, x_2: jnp.ndarray, bandwidth: ) -> jnp.ndarray`

Where `x_1` and `x_2` are the positions to calculate the kernel between, and `kernel_kwargs` are any additional arguments to the kernel function (e.g. bandwidth). The kernel must be normalized with volume 1. It should be a pure function written in JAX so it is compatible with jax.vmap. 

"""

import jax.numpy as jnp
from jax import jit
from kalmax.utils import gaussian_norm_const


def gaussian_kernel(x1 : jnp.ndarray,
                    x2 : jnp.ndarray, 
                    bandwidth : float,
                    ) -> jnp.ndarray:
    """Calculates the gaussian kernel between two points x1 and x2 with covariance

    Parameters
    ----------

    x1: (D,) jnp.ndarray
        The first position
    x2: (D,) jnp.ndarray
        The second position
    bandwidth: float
        The bandwidth of the kernel
    
    Returns
    -------
    kernel: float
        The probability density at x
    """
    assert x1.ndim == 1
    assert x2.ndim == 1
    assert x1.shape[0] == x2.shape[0]
    D = x1.shape[0]    

    covariance = jnp.eye(D) * bandwidth ** 2
    x = x1 - x2
    norm_const = gaussian_norm_const(covariance)
    kernel = norm_const * jnp.exp(-0.5 * jnp.sum(x @ jnp.linalg.inv(covariance) * x))
    return kernel


def laplacian_kernel(x1 : jnp.ndarray,
                     x2 : jnp.ndarray,
                     bandwidth : float) -> jnp.ndarray:
        """Calculates the laplacian kernel between two points x1 and x2 with bandwidth h
        
        Parameters
        ----------
        
        x1: (D,) jnp.ndarray
            The first position
        x2: (D,) jnp.ndarray
            The second position
        bandwidth: float
            The bandwidth of the kernel
        
        Returns
        -------
        kernel: float
            The probability density at x
        """
        assert x1.ndim == 1
        assert x2.ndim == 1
        assert x1.shape[0] == x2.shape[0]
        
        x = jnp.linalg.norm(x1 - x2)
        kernel = jnp.exp(-x / bandwidth)
        return kernel


def uniform_kernel(x1 : jnp.ndarray,
                   x2 : jnp.ndarray,
                   bandwidth : float) -> jnp.ndarray:
    """Calculates the uniform kernel between two points x1 and x2 with bandwidth h
    
    Parameters
    ----------
    
    x1: (D,) jnp.ndarray
        The first position
    x2: (D,) jnp.ndarray
        The second position
    bandwidth: float
        The bandwidth of the kernel
     
    Returns
    -------
    kernel: float
        The probability density at x
    """
    assert x1.ndim == 1
    assert x2.ndim == 1
    assert x1.shape[0] == x2.shape[0]
    
    x = jnp.linalg.norm(x1 - x2)
    kernel = jnp.where(x < bandwidth, 1 / bandwidth, 0)
    return kernel

def epanechnikov_kernel(x1 : jnp.ndarray,
                        x2 : jnp.ndarray,
                        bandwidth : float) -> jnp.ndarray:
    """Calculates the epanechnikov kernel between two points x1 and x2 with bandwidth h
    
    Parameters
    ----------
    
    x1: (D,) jnp.ndarray
        The first position
    x2: (D,) jnp.ndarray
        The second position
    bandwidth: float
        The bandwidth of the kernel
     
    Returns
    -------
    kernel: float
        The probability density at x
    """
    assert x1.ndim == 1
    assert x2.ndim == 1
    assert x1.shape[0] == x2.shape[0]
    
    x = jnp.linalg.norm(x1 - x2)
    kernel = jnp.where(x < bandwidth, 3 / (4 * bandwidth) * (1 - (x / bandwidth) ** 2), 0)
    return kernel

def triangular_kernel(x1 : jnp.ndarray,
                      x2 : jnp.ndarray,
                      bandwidth : float) -> jnp.ndarray:
        """Calculates the triangular kernel between two points x1 and x2 with bandwidth h
        
        Parameters
        ----------
        
        x1: (D,) jnp.ndarray
            The first position
        x2: (D,) jnp.ndarray
            The second position
        bandwidth: float
            The bandwidth of the kernel
         
        Returns
        -------
        kernel: float
            The probability density at x
        """
        assert x1.ndim == 1
        assert x2.ndim == 1
        assert x1.shape[0] == x2.shape[0]
        
        x = jnp.linalg.norm(x1 - x2)
        kernel = jnp.where(x < bandwidth, 1 - x / bandwidth, 0)
        return kernel


import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="kalmax",  
    version="0.1.0", 
    author="Tom George",
    author_email="tomgeorge1@btinternet.com",
    description="Kalman based neural decoding in Jax",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/TomGeorge1234/KalMax", 
    packages=setuptools.find_packages(), # finds all packages inside the parent directory containing a __init__.py file
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
    install_requires=[
        'jax',
    ],
    extras_require={
        'demo': ['matplotlib', 'tqdm', 'ratinabox','pykalman'],
    }
)



import jax 
import jax.numpy as jnp
import tqdm as tqdm

def gaussian_pdf(x : jnp.ndarray,
                 mu : jnp.ndarray, 
                 sigma : jnp.ndarray,) -> jnp.ndarray:
    """ Calculates the gaussian pdf of a multivariate normal distribution of mean mu and covariance sigma at x

    Parameters
    ----------

    x: (D,) array
        The position at which to evaluate the pdf
    mu: (D,) array
        The mean of the distribution
    sigma: (D, D) array
        The covariance of the distribution
    
    Returns
    -------
    pdf: float
        The probability density at x
    """
    assert x.ndim == 1
    assert mu.ndim == 1
    assert sigma.ndim == 2
    assert x.shape[0] == mu.shape[0]
    assert x.shape[0] == sigma.shape[0]
    assert sigma.shape[0] == sigma.shape[1]

    x = x - mu
    norm_const = gaussian_norm_const(sigma)
    return norm_const * jnp.exp(-0.5 * jnp.sum(x @ jnp.linalg.inv(sigma) * x, axis=-1))

def log_gaussian_pdf(x : jnp.ndarray,
                     mu : jnp.ndarray,
                     sigma : jnp.ndarray,) -> jnp.ndarray:
    """ Calculates the log of the gaussian pdf of a multivariate normal distribution of mean mu and covariance sigma at x

    Parameters
    ----------
    x: (D,) array
        The position at which to evaluate the pdf
    mu: (D,) array
        The mean of the distribution
    sigma: (D, D) array
        The covariance of the distribution

    Returns 
    -------
    log_pdf: float
        The log probability density at x
    """
    assert x.ndim == 1
    assert mu.ndim == 1
    assert sigma.ndim == 2
    assert x.shape[0] == mu.shape[0]
    assert x.shape[0] == sigma.shape[0]
    assert sigma.shape[0] == sigma.shape[1]

    x = x - mu
    norm_const = gaussian_norm_const(sigma)
    return jnp.log(norm_const) - 0.5 * jnp.sum(x @ jnp.linalg.inv(sigma) * x)



def gaussian_norm_const(sigma : jnp.ndarray) -> jnp.ndarray:
    """Calculates the normalizing constant of a multivariate normal distribution with covariance sigma

    Parameters
    ----------
    sigma: jnp.ndarray, shape (D, D)
        The covariance matrix of the distribution

    Returns
    -------
    norm_const: jnp.ndarray, shape (1,)
        The normalizing constant
    """
    assert sigma.ndim == 2
    D = sigma.shape[0]
    return 1 / jnp.sqrt((2 * jnp.pi) ** D * jnp.linalg.det(sigma))

def fit_gaussian(x, likelihood):
    """Fits a multivariate-Gaussian to the likelihood function P(spikes | x) in x-space.
    
    Parameters
    ----------
    x : jnp.ndarray, shape (N_bins,D)
        The position bins in which the likelihood is calculated
    likelihood : jnp.ndarray, shape (N_bins,)
        The combined likelihood (not log-likelihood) of the neurons firing at each position bin
        
    Returns
    -------
    mu : jnp.ndarray, shape (D,)
        The mean of the Gaussian
    mode : jnp.ndarray, shape (D,)
        The mode of the Gaussian
    covariance : jnp.ndarray, shape (D, D)
        The covariance of the Gaussian    
    """
    assert x.ndim == 2
    assert likelihood.ndim == 1
    assert x.shape[0] == likelihood.shape[0]
    
    mu = (x.T @ likelihood) / likelihood.sum()
    mode = x[jnp.argmax(likelihood)]
    covariance = ((x - mu) * likelihood[:, None]).T @ (x - mu) / likelihood.sum()
    return mu, mode, covariance


# Like fit_gaussian, but accepts likelihoods of shape (T, N_bins)
# returns means, modes and covariances of shape (T, D), (T, D), (T, D, D)
fit_gaussian_vmap = jax.vmap(fit_gaussian, in_axes=(None, 0)) 



def make_simulated_dataset(time_mins = 60, n_cells = 100, firing_rate = 10, random_seed=None, **kwargs):
    """Makes a simulated dataset for an agent randomly foraging a 1 m square box. Data generated with the RatInABox package and defaults ot 50 place cells. Returns the data as jax arrays.
    
    Parameters
    ----------
    time_mins: int
        The number of minutes to simulate the agent for
    kwargs: dict
        Additional arguments to pass to the RatInABox simulation
    
    Returns
    -------
    time: jnp.ndarray, shape (N,)
        The time points of the simulation
    position: jnp.ndarray, shape (N, dims)
        The position of the agent at each time point
    spikes: jnp.ndarray, shape (N, N_cells)
        The spikes of the 50 place cells at each time point
    """

    from ratinabox.Environment import Environment 
    from ratinabox.Agent import Agent 
    from ratinabox.Neurons import PlaceCells

    if random_seed is not None:
        import numpy as np
        np.random.seed(random_seed)
    
    env_params = kwargs.get('env_params', {})
    agent_params = kwargs.get('agent_params', {'dt':0.1})
    place_cell_params = kwargs.get('place_cell_params', {'n':n_cells, 'max_fr':firing_rate,'widths':0.1})

    env = Environment(params=env_params)
    agent = Agent(env, params=agent_params)
    place_cells = PlaceCells(agent, params=place_cell_params)
    

    for i in tqdm.tqdm(range(int(60 * time_mins / agent.dt))):
        agent.update()
        place_cells.update()

    time  = jnp.array(agent.history['t'])
    position = jnp.array(agent.history['pos'])
    spikes = jnp.array(place_cells.history['spikes'])

    return time, position, spikes

